package expressionprog;

import java.util.ArrayList;

/**
 * This class provides one static public method for parsing expressions.
 * The method produces a program for a stack machine in the form of
 * a list of ProgramItems.  Expressions can include:
 *     floating point numbers;
 *     variables, with names starting with a letter and containing letters and digits;
 *     binary operators +, -, *, /, and ^, where x^y means x raised to the power y;
 *     unary - operator, as in "-x";
 *     functions of one variable, followed by an argument enclosed in parentheses;
 *     assignment operations of the form variable = expression;
 *     the special variables #, #1, #2, #3, ....
 * Variables are distinguished from function names only by the "(" that follows
 * a function name; any name that is not followed by "(" is considered to be
 * a variable.  The parser assigns no meaning to variable or function names;
 * that is up to the part of the program that evaluates the expression.  The
 * special variables #, #1, #2, ..., are meant to represent "previous values."
 * That is they are the values of previously evaluated expressions.  # is meant
 * to refer to the value of the expression that was evaluated just before
 * the one that contains the #.  #1 is meant to refer to the first expressions
 * that was evaluate, #2 to the second expression, #3 to the third expression,
 * and so on.  However, the parser simply creates a ProgramItem of type
 * PreviousValue to represent the special variable.  It is up to the program
 * that uses the parser to deal with their values.
 */
public class Parser {

	/**
	 * Parses an expression and returns a list of ProgramItems.  The list
	 * is a program for a stack machine, and executing the program will
	 * evaluate the expression and leave the result on the stack.
	 * Note that the parser does NOT check that variables and functions
	 * that are used in the expression are actually defined.  The
	 * code that evaluates the expression will have to deal with that
	 * issue.
	 * @throws IllegalArgumentException if the string does not contain
	 *     a single, legal expression.
	 */
	public static ArrayList<ProgramItem> parse(String expression) {
		if (expression == null)
			throw new IllegalArgumentException("Can't parse a null string.");
		expression = expression.trim();
		if (expression.length() == 0)
			throw new IllegalArgumentException("Can't parse an empty string.");
		program = new ArrayList<ProgramItem>();
		input = expression;
		pos = 0;
		parseExpression();
		skip();
		if (pos < input.length())
			throw new IllegalArgumentException("Extra data found in input after end of expression \"" 
					+ input.substring(0,pos).trim() + "\"");
		return program;
	}

	// -------------- Private implementation part ---------------------

	private static ArrayList<ProgramItem> program;
	private static String input;
	private static int pos;

	private static void skip() {  // skip past whitespace in input
		while (pos < input.length() && Character.isWhitespace(input.charAt(pos)))
			pos++;
	}

	private static char peek() {  // look at next char in input
		if (pos >= input.length())
			return 0;
		else
			return input.charAt(pos);
	}

	private static char next() {  // look at and remove next char in input
		if (pos == input.length())
			throw new IllegalArgumentException("Unexpected end of input while parsing.");
		return input.charAt(pos++);
	}

	private static void parseExpression() {
		boolean neg = false;
		skip();
		if (peek() == '+' || peek() == '-') {
			if (next() == '-')
				neg = true;
		}
		parseTerm();
		if (neg)
			program.add(new UnaryMinus());
		skip();
		while (peek() == '+' || peek() == '-') {
			char op = next();
			parseTerm();
			program.add(new BinOp(op));
			skip();
		}
	}

	private static void parseTerm() {
		parseFactor();
		skip();
		while (peek() == '*' || peek() == '/') {
			char op = next();
			parseFactor();
			program.add(new BinOp(op));
			skip();
		}
	}

	private static void parseFactor() {
		parsePrimary();
		skip();
		while (peek() == '^') {
			next();
			parsePrimary();
			program.add(new BinOp('^'));
			skip();
		}
	}

	private static void parsePrimary() {
		skip();
		char ch = peek();
		if (Character.isLetter(ch))
			parseWord();
		else if ( (ch >= '0' && ch <= '9') || ch == '.' )
			parseNumber();
		else if (ch == '(') {
			next();
			parseExpression();
			skip();
			if (peek() == ')')
				next();
			else {
				if (peek() == 0)
					throw new IllegalArgumentException("Missing right parenthesis at end of expression.");
				else
					throw new IllegalArgumentException("Expecting a right parenthese, found \"" + peek() + "\"");
			}
		}
		else if (ch == '#') {
			next();
			String name = "#";
			while (peek() >= '1' && peek() <= '9')
				name = name + next();
			skip();
			if (peek() == '=')
				throw new IllegalArgumentException("Can't assign a value to the special variable " + name);
			program.add( new Variable( name ));
		}
		else if (ch == 0)
			throw new IllegalArgumentException("Expecting a number, variable, etc, but hit the end of input.");
		else
			throw new IllegalArgumentException("Found illegal or unexpected character \"" + peek() + "\" in input.");
	}

	private static void parseNumber() {
		String num = new String();
		while (peek() >= '0' && peek() <= '9')
			num += next();
		if (peek() == '.') {
			num += next();
			while (peek() >= '0' && peek() <= '9')
				num += next();
		}
		if (peek() == 'e' || peek() == 'E') {
			num += next();
			if (peek() == '-' || peek() == '+')
				num += next();
			while (peek() >= '0' && peek() <= '9')
				num += next();
		}
		double x;
		try {
			x = Double.parseDouble(num);
			if (Double.isInfinite(x))
				throw new IllegalArgumentException("Number \"" + num + "\" is too big!");
		}
		catch (NumberFormatException e) {
			throw new IllegalArgumentException("Illegal number \"" + num + "\" found in input.");
		}
		program.add( new Number(x) );
	}

	private static void parseWord() {
		String name = "";
		while (Character.isLetter(peek()) || Character.isDigit(peek()))
			name += next();
		skip();
		if (peek() == '=') {
			next();
			parseExpression();
			program.add(new Assign(name) );
		}
		else if (peek() != '(') {
			program.add( new Variable(name) );
		}
		else {
			next();
			parseExpression();
			skip();
			if (peek() == ')') {
				next();
				program.add( new Function(name) );
			}
			else {
				throw new IllegalArgumentException("Missing right parenthesis to end the argument of function " + name);
			}
		}
	}

}
